/*!
 * \file bitMapMarker.hpp
 * \brief This file is used to store and control of bit map.
 * Copyright (C) 2011-2013 Nippon Telegraph and Telephone Corporation
 *
 * [License]
 *
 * $id$
 */

#ifndef _BITMAPMARKER_HPP
#define _BITMAPMARKER_HPP

#include <stddef.h>

/*!
 * \brief This class is stored bit express flag in pointer range.
 */
class TBitMapMarker {
  public:
    /*!
     * \brief TBitMapMarker constructor.
     * \param startAddr [in] Start address of Java heap.
     * \param size      [in] Max Java heap size.
     */
    TBitMapMarker(void *startAddr, size_t size);
    /*!
     * \brief TBitMapMarker destructor.
     */
    virtual ~TBitMapMarker();
    
    /*!
     * \brief Check address which is covered by this bitmap.
     * \param addr [in] Oop address.
     * \return Designated pointer is existing in bitmap.
     */
    inline bool isInZone(void *addr) {
      return (this->beginAddr <= addr) && (this->endAddr >= addr);
    }
    
    /*!
     * \brief Mark GC-marked address in this bitmap.
     * \param addr [in] Oop address.
     */
    void setMark(void *addr);
    
    /*!
     * \brief Get marked flag of designated pointer.
     * \param addr [in] Targer pointer.
     * \return Designated pointer is marked.
     */
    bool isMarked(void *addr);
    
    /*!
     * \brief Check address which is already marked and set mark.
     * \param addr [in] Oop address.
     * \return Designated pointer is marked.
     */
    bool checkAndMark(void *addr);
    
    /*!
     * \brief Clear bitmap flag.
     */
    void clear(void);
    
  protected:
    /*!
     * \brief Get a byte data and mask expressing target pointer.
     * \param addr  [in]  Oop address.
     * \param block [out] A byte data including bit flag.
     *                    The bit flag is expressing target pointer.
     * \param mask  [out] A bitmask for getting only flag of target pointer.
     */
    virtual void getBlockAndMask(void *addr, ptrdiff_t **block, ptrdiff_t *mask);
    
  private:
    /*!
     * \brief Begin of Java heap.
     */
    void *beginAddr;
    /*!
     * \brief End of Java heap.
     */
    void *endAddr;
    /*!
     * \brief Real allocated bitmap range size for mmap.
     */
    size_t bitmapSize;
    
    /*!
     * \brief Pointer of memory stored bitmap flag.
     */
    void *bitmapAddr;
};

#endif // _BITMAPMARKER_HPP
