/*
 * SnapShotHeader.java
 * Created on 2011/10/13
 * RCSID: $Id:$
 *
 * Copyright (C) 2011-2013 Nippon Telegraph and Telephone Corporation
 *
 */

package jp.co.ntt.oss.heapstats.container;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.nio.ByteOrder;
import java.util.Date;

/**
 * implements {@link Serializable} <br>
 * <br>
 * Contains the header information of the snapshot.Also, write to the temporary
 * file, do the reading.
 */
public class SnapShotHeader implements
                           Serializable {
    /** serialVersionUID. */
    static final long serialVersionUID = -539015033687122109L;

    /** SnapShot Case GC. */
    private static final int SNAPSHOT_CAUSE_GC = 1;

    /** SnapShot Case DUMP REQUEST. */
    private static final int SNAPSHOT_CAUSE_DATA_DUMP_REQUEST = 2;

    /** SnapShot Case INTERVAL. */
    private static final int SNAPSHOT_CAUSE_INTERVAL = 3;

    /** Byte order of the agent Dump file. */
    private ByteOrder byteOrderMark;

    /** Time the snapshot was taken. */
    private Date snapShotDate;

    /** Number of Objects. */
    private long numEntries;

    /** SnapShot Cause. */
    private int cause;

    /** GC Cause. */
    private String gcCause;

    /** Full GC Count. */
    private long fullCount;

    /** Young GC Count. */
    private long yngCount;

    /** GC Time. */
    private long gcTime;

    /** New Heap Size. */
    private long newHeap;

    /** old Heap Size. */
    private long oldHeap;

    /** Total Heap Size. */
    private long totalSize;

    /**
     * Creates a SnapShotHeader.
     */
    public SnapShotHeader() {
        byteOrderMark = ByteOrder.nativeOrder();
        snapShotDate = null;
        numEntries = 0;
        cause = 0;
        fullCount = 0;
        yngCount = 0;
        gcTime = 0;
        newHeap = 0;
        oldHeap = 0;
        totalSize = 0;
    }

    /**
     * Set the byte order.
     *
     * @param value ByteOrder
     */
    public final void setByteOrderMark(final ByteOrder value) {
        byteOrderMark = value;
    }

    /**
     * Return the byte order.
     *
     * @return Return the byte order
     */
    public final ByteOrder getByteOrderMark() {
        return byteOrderMark;
    }

    /**
     * Set the date and time the snapshot was taken.
     *
     * @param value Time the snapshot was taken
     */
    public final void setSnapShotDate(final Date value) {
        snapShotDate = value;
    }

    /**
     * Set the date and time the snapshot was taken.
     *
     * @param value Time the snapshot was taken
     */
    public final void setSnapShotDateAsLong(final long value) {
        snapShotDate = new Date(value);
    }

    /**
     * To get the SnapShot Date.
     *
     * @return Return the SnapShot Date
     */
    public final Date getSnapShotDate() {
        return snapShotDate;
    }

    /**
     * To set the Number of Object entries.
     *
     * @param value the Number of Object entries
     */
    public final void setNumEntries(final long value) {
        numEntries = value;
    }

    /**
     * To get the Number of Object entries.
     *
     * @return Return the Number of Object entries
     */
    public final long getNumEntries() {
        return numEntries;
    }

    /**
     * Set the source snapshot was taken.
     *
     * @param value Cause the snapshot was taken
     */
    public final void setCause(final int value) {
        cause = value;
    }

    /**
     * Return the Cause the snapshot was taken.
     *
     * @return Return the Cause the snapshot was taken
     */
    public final int getCause() {
        return cause;
    }

    /**
     * @return Return the GC Cause
     */
    public final String getCauseString() {
        switch (cause) {
        case SNAPSHOT_CAUSE_GC:
            return "GC";
        case SNAPSHOT_CAUSE_DATA_DUMP_REQUEST:
            return "DataDumpRequest";
        case SNAPSHOT_CAUSE_INTERVAL:
            return "Interval";
        default:
            return "Unknown";
        }
    }

    /**
     * Return Cause the GC.
     *
     * @return Return Cause the GC
     */
    public final String getGcCause() {
        return gcCause;
    }

    /**
     * Sets the cause of the GC.
     *
     * @param value Cause the GC
     */
    public final void setGcCause(final String value) {
        gcCause = value;
    }

    /**
     * Return the number of occurrences of FullGC.
     *
     * @return Return the number of occurrences of FullGC
     */
    public final long getFullCount() {
        return fullCount;
    }

    /**
     * Set the number of occurrences of FullGC.
     *
     * @param value number of occurrences of FullGC
     */
    public final void setFullCount(final long value) {
        fullCount = value;
    }

    /**
     * Return the number of occurrences of YoungGC.
     *
     * @return Return the number of occurrences of YoungGC
     */
    public final long getYngCount() {
        return yngCount;
    }

    /**
     * Set the number of occurrences of YoungGC.
     *
     * @param value number of occurrences of YoungGC
     */
    public final void setYngCount(final long value) {
        yngCount = value;
    }

    /**
     * Return the processing time of the GC.
     *
     * @return Return the processing time of the GC
     */
    public final long getGcTime() {
        return gcTime;
    }

    /**
     * Set the processing time of the GC.
     *
     * @param value the processing time of the GC
     */
    public final void setGcTime(final long value) {
        gcTime = value;
    }

    /**
     * Return the New heap space usage.
     *
     * @return Return the New heap space usage
     */
    public final long getNewHeap() {
        return newHeap;
    }

    /**
     * Set the heap space usage New.
     *
     * @param value the heap space usage New
     */
    public final void setNewHeap(final long value) {
        newHeap = value;
    }

    /**
     * Return the Old heap space usage.
     *
     * @return Return the Old heap space usage
     */
    public final long getOldHeap() {
        return oldHeap;
    }

    /**
     * Set the heap space usage Old.
     *
     * @param value the heap space usage Old
     */
    public final void setOldHeap(final long value) {
        oldHeap = value;
    }

    /**
     * Return the total heap usage.
     *
     * @return Return the total heap usage
     */
    public final long getTotalSize() {
        return totalSize;
    }

    /**
     * Sets the total heap usage.
     *
     * @param value the total heap usage
     */
    public final void setTotalSize(final long value) {
        totalSize = value;
    }

    @Override
    public final String toString() {
        StringBuffer buf = new StringBuffer();

        buf.append("SnapShot at ");
        buf.append(snapShotDate);
        buf.append(", ");
        buf.append(numEntries);
        buf.append(" entries, ");
        buf.append(byteOrderMark);
        buf.append(" Caused by ");
        buf.append(getCauseString());
        buf.append(", GC Caused by ");
        buf.append(gcCause);
        buf.append(", Full GC ");
        buf.append(fullCount);
        buf.append(", Young GC ");
        buf.append(yngCount);
        buf.append(", GC Time at ");
        buf.append(gcTime);
        buf.append(", New Heap Usage ");
        buf.append(newHeap);
        buf.append(" byte ");
        buf.append(", Old Heap Usage ");
        buf.append(oldHeap);
        buf.append(" byte ");
        buf.append(", Total Heap Usage ");
        buf.append(totalSize);
        buf.append(" byte ");

        return buf.toString();
    }

    /**
     * Reads the header information from the temporary file.
     *
     * @param stream the temporary file input stream
     * @throws IOException Any exception thrown by the underlying OutputStream.
     * @throws ClassNotFoundException Class of a serialized object cannot be
     *         found.
     */
    private void readObject(final ObjectInputStream stream) throws IOException,
            ClassNotFoundException {
        snapShotDate = (Date) stream.readObject();
        numEntries = stream.readLong();
        if (stream.readBoolean()) {
            byteOrderMark = ByteOrder.BIG_ENDIAN;
        } else {
            byteOrderMark = ByteOrder.LITTLE_ENDIAN;
        }
        cause = stream.readInt();

        // GC Case
        int length = stream.readInt();
        StringBuilder tmpGccause = new StringBuilder(length);
        for (int i = 0; i < length; i++) {
            tmpGccause.append(stream.readChar());
        }

        gcCause = tmpGccause.toString();

        fullCount = stream.readLong();
        yngCount = stream.readLong();
        gcTime = stream.readLong();
        newHeap = stream.readLong();
        oldHeap = stream.readLong();
        totalSize = stream.readLong();
    }

    /**
     * Writes header information to a temporary file.
     *
     * @param stream the temporary file output stream
     * @throws IOException Any exception thrown by the underlying OutputStream.
     */
    private void writeObject(final ObjectOutputStream stream)
            throws IOException {
        stream.writeObject(snapShotDate);
        stream.writeLong(numEntries);
        stream.writeBoolean(byteOrderMark == ByteOrder.BIG_ENDIAN);
        stream.writeInt(cause);
        stream.writeInt(gcCause.length());
        stream.writeChars(gcCause);
        stream.writeLong(fullCount);
        stream.writeLong(yngCount);
        stream.writeLong(gcTime);
        stream.writeLong(newHeap);
        stream.writeLong(oldHeap);
        stream.writeLong(totalSize);
    }

}
